import json
from pathlib import Path
import logging
import subprocess
import importlib.metadata
import sys
from rockai.server.utils import get_dependencies, load_predictor_class
from rockai.command.openapi_schema import get_openapi_json

logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("docker").setLevel(logging.WARNING)


def remove_some_libs(file_name, lib_to_be_deleted):
    try:
        # Open the file in read mode
        with open(file_name, "r") as file:
            lines = file.readlines()

        # Open the file in write mode
        with open(file_name, "w") as file:
            for line in lines:
                if lib_to_be_deleted not in line:
                    file.write(line)

        print("Successfully removed {} from {}".format(lib_to_be_deleted, file_name))
    except FileNotFoundError:
        print("{} not found".format(file_name))


def get_tensorflow_version(filename: str) -> str:
    """
    Reads a pip's requirements.txt file and returns the TensorFlow version specified in it.
    """
    try:
        with open(filename, "r") as file:
            for line in file:
                if line.startswith("tensorflow=="):
                    # Split the line at '==' to get the package and version
                    parts = line.strip().split("==")
                    if len(parts) == 2:
                        return parts[1]
                    else:
                        return "No specific version specified"
        return None
    except FileNotFoundError:
        return "File not found"


def run_command_and_get_output(file_name):
    # Run the command and capture the output
    output = get_openapi_json(file_name)

    return "'{}'".format(
        str(json.dumps(json.dumps(json.loads(output)), separators=(",", ":")))
    )


def build_docker_image_without_configuration(
    image_name: str,
    predictor_file_name: str,
    port: int,
    is_using_gpu: bool,
    platform="linux/amd64",
    is_dry_run: bool = False,
    upload_url: str = "",
    is_using_mirror: bool = True,
):
    python_version = (
        f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
    )
    base_image = (
        "nvidia/cuda:11.8.0-cudnn8-devel-ubuntu22.04"
        if is_using_gpu
        else "python:{}".format(python_version)
    )

    docker_list = []
    docker_list.append(add_base(base_image))
    docker_list.append(add_env("DEBIAN_FRONTEND=noninteractive"))
    docker_list.append(add_run("export HF_ENDPOINT=https://hf-mirror.com"))
    docker_list.append(add_env("HF_HUB_ENABLE_HF_TRANSFER=1"))
    docker_list.append(add_work_dir("/src"))
    docker_list.append(copy_files(".", "/src"))
    if is_using_gpu:  # gpu support
        docker_list.append(
            add_run(
                "sed -i 's/http:\/\/archive.ubuntu.com\/ubuntu\//http:\/\/mirrors.tuna.tsinghua.edu.cn\/ubuntu\//g' /etc/apt/sources.list"
            )
        )
        docker_list.append(add_run("apt clean"))
        docker_list.append(add_run("apt-get clean"))
        docker_list.append(add_run("apt update"))
        docker_list.append(add_run("apt-get update"))
        docker_list.append(add_env('PATH="/root/.pyenv/shims:/root/.pyenv/bin:$PATH"'))
        docker_list.append(
            add_run(
                """--mount=type=cache,target=/var/cache/apt apt-get update -qq && apt-get install -qqy --no-install-recommends \
        make \
        build-essential \
        libssl-dev \
        zlib1g-dev \
        libbz2-dev \
        libreadline-dev \
        libsqlite3-dev \
        wget \
        curl \
        llvm \
        libncurses5-dev \
        libncursesw5-dev \
        xz-utils \
        tk-dev \
        libffi-dev \
        liblzma-dev \
        git \
        ca-certificates \
        && rm -rf /var/lib/apt/lists/*
    """
            )
        )
        docker_list.append(
            add_run(
                """curl -s -S -L https://raw.githubusercontent.com/pyenv/pyenv-installer/master/bin/pyenv-installer | bash && \
	git clone https://github.com/momo-lab/pyenv-install-latest.git "$(pyenv root)"/plugins/pyenv-install-latest && \
	pyenv install-latest {} && \
	pyenv global $(pyenv install-latest --print {}) && \
	pip install "wheel<1"
""".format(
                    python_version,
                    python_version,
                )
            )
        )
    else:  # cpu only
        docker_list.append(add_run("apt update"))
        docker_list.append(add_run("apt-get update"))

    docker_list.append(add_expose(port))
    docker_list.append(
        add_run(f"pip install rockai=={importlib.metadata.version('rockai')}")
    )
    pred = load_predictor_class(predictor_file_name)
    ## install python library if any
    py_libs = get_dependencies(pred, "requirement_dependency")
    print(py_libs)
    for item in py_libs:

        docker_list.append(
            add_run(
                f"pip install {item} -i https://pypi.tuna.tsinghua.edu.cn/simple"
                if is_using_mirror
                else f"pip install {item}"
            )
        )
    docker_list.append(
        add_run(
            "pip install huggingface_hub hf_transfer -i https://pypi.tuna.tsinghua.edu.cn/simple"
            if is_using_mirror
            else "pip install huggingface_hub hf_transfer"
        )
    )
    ## install system library if any
    docker_list.append(add_run('apt update'))
    system_libs = get_dependencies(pred, "system_dependency")
    print(system_libs)
    for item in system_libs:
        docker_list.append(add_run(f"apt install -y {item}"))
    ## run custom cmd if any
    custom_cmds = get_dependencies(pred, "custom_cmd")
    for item in custom_cmds:
        docker_list.append(add_run(item))
        
    docker_list.append(
        add_labels(
            "run.cog.openapi_schema", run_command_and_get_output(predictor_file_name)
        )
    )
    docker_list.append(
        add_cmd(
            "export HF_HUB_ENABLE_HF_TRANSFER=1 && export HF_ENDPOINT=https://hf-mirror.com && rockai start --port {} --file {} --upload-url {}".format(
                port, predictor_file_name, upload_url
            )
        )
    )

    save_docker_file(docker_list)
    if not is_dry_run:
        subprocess.run(
            [
                "docker",
                "build",
                "--platform",
                platform,
                "-t",
                image_name,
                "-f",
                Path.cwd() / ".rock_temp/Dockerfile",
                Path.cwd(),
            ]
        )


def add_base(base):
    return "FROM {}\n".format(base)


def add_cmd(cmd_list):
    return "CMD {}\n".format(cmd_list)


def add_expose(port):
    return "EXPOSE {}\n".format(port)


def add_work_dir(dir):
    return "WORKDIR {}\n".format(dir)


def add_run(cmd):
    return "RUN {}\n".format(cmd)


def copy_files(src, dest):
    return "COPY {} {}\n".format(src, dest)


def add_env(env):
    return "ENV {}\n".format(env)


def add_labels(key, value):
    return "LABEL {}={}\n".format(key, value)


def save_docker_file(cmd_list):
    result = "".join(cmd_list)
    # Define the directory path
    directory_path = Path(str(Path.cwd() / ".rock_temp"))

    # Check if the directory exists
    if not directory_path.exists():
        # If it does not exist, create the directory
        directory_path.mkdir(parents=True, exist_ok=True)
        print(f"Directory created: {directory_path}")
    else:
        print(f"Directory already exists: {directory_path}")
    try:
        with open(Path.cwd() / ".rock_temp/Dockerfile", "w") as file:
            file.write(result)
    except Exception as e:
        print("An error occurred while writing to the file. Error: ", str(e))



import typer
from rockai.server.http import start_server
from rockai.server.utils import is_valid_name
from pathlib import Path
from typing_extensions import Annotated
from rockai.docker.docker_util import build_docker_image_without_configuration
import requests
import subprocess
import logging
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich import print
import getpass

logger = logging.getLogger()
logging.basicConfig(level=logging.DEBUG)

app = typer.Typer()

APP_NAME = "rockai"


def download_file(url: str, save_path: str):
    try:
        # Send a GET request to the URL
        response = requests.get(url, stream=True)
        # Check if the request was successful
        response.raise_for_status()

        # Open the file in binary write mode and write the content
        with open(save_path, "wb") as file:
            for chunk in response.iter_content(chunk_size=8192):
                file.write(chunk)

    except requests.exceptions.RequestException as e:
        print(f"Failed to download the file: {str(e)}")


@app.command()
def init():
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        progress.add_task(description="Initiating, please wait...", total=None)
        download_file(
            "https://rockai-web-resouce-bucket.s3.ap-northeast-1.amazonaws.com/predictor.py",
            Path.cwd() / "predictor.py",
        )
        download_file(
            "https://rockai-web-resouce-bucket.s3.ap-northeast-1.amazonaws.com/dockerignore",
            Path.cwd() / ".dockerignore",
        )
        print("Init success!")


@app.command(name="build")
def build(
    port: Annotated[int, typer.Option(help="Port of the API server")] = 8000,
    name: Annotated[
        str,
        typer.Argument(
            help="Image name of the docker container --> Example: r.18h.online/jian-yang/hotdog-detector"
        ),
    ] = None,
    file: Annotated[
        str, typer.Option(help="Path to predictor.py file,default is predictor.py")
    ] = "predictor.py",
    gpu: Annotated[bool, typer.Option(help="Is using gpu")] = True,
    platform: Annotated[
        str,
        typer.Option(
            help="Docker image supported platform, `linux/amd64` by default, you can also change it to other platform like `linux/arm64/v8` if you are using MAC M2 chip"
        ),
    ] = "linux/amd64",
    dry_run: Annotated[
        bool, typer.Option(help="Generate Dockerfile without building the image")
    ] = False,
    upload_url: Annotated[
        str,
        typer.Option(
            help="Url for upload model output, Example:-> 'https://api.rockai.online/v1/get_presign_url?file_name=output.png' "
        ),
    ] = "https://api.rockai.online/v1/get_presign_url",
    mirror: Annotated[
        bool, typer.Option(help="is using mirror to download python packages, default mirror is Tsinghua University mirror")
    ] = True
):
    """
    Build the image
    """

    if name is not None:
        if name.startswith("r.18h.online"):
            if is_valid_name(name.split("/")[-1]):
                build_docker_image_without_configuration(
                    name, file, port, gpu, platform, dry_run, upload_url,mirror
                )
            else:
                print(
                    f"Invalid model name '{name}'. Model name should start with a letter, contain only alphanumeric characters and hyphens, and be less than or equal to 140 characters long."
                )
        else:
            build_docker_image_without_configuration(
                name, file, port, gpu, platform, dry_run, upload_url,mirror
            )


@app.command()
def start(
    auth: Annotated[
        str, typer.Option(help="Bearer auth token of the API server")
    ] = None,
    port: Annotated[int, typer.Option(help="Port of the API server")] = 8000,
    file: Annotated[
        str, typer.Option(help="Path to predictor.py file,default is predictor.py")
    ] = "predictor.py",
    upload_url: Annotated[
        str,
        typer.Option(
            help="Url for upload model output, Example:-> 'https://api.rockai.online/v1/get_presign_url?file_name=output.png' "
        ),
    ] = "https://api.rockai.online/v1/get_presign_url",
):
    """
    start local development server
    """
    start_server(file, port, auth, upload_url)


@app.command("push")
def push_model(
    name: Annotated[
        str,
        typer.Argument(
            help="name of the image you want to push to rockai server,  --> Example: 'r.18h.online/jian-yang/hotdog-detector' "
        ),
    ]
):
    """
    Push the model to the RockAI platform
    """
    subprocess.run(["docker", "image", "push", "{}".format(name)])


@app.command(name="login")
def login_to_docker(
    api_token: Annotated[str, typer.Option(help="Your API token")] = None,
    debug: Annotated[bool, typer.Option(help="Enable debug mode")] = False,
):
    """
    Login to the RockAI Docker registry
    """
    if api_token is None:
        api_token = getpass.getpass("Paste your API token:")
        print(api_token)
    url = "https://api.rockai.online/v1/user/docker_token"

    if debug:
        print("Debug mode is enabled")
        url = "https://api-dev.rockai.online/v1/user/docker_token"

    headers = {"Authorization": "Bearer {}".format(api_token)}
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        progress.add_task(description="Logging in, please wait...", total=None)
        response = requests.get(url, headers=headers)
        response.raise_for_status()
        subprocess.run(
            [
                "docker",
                "login",
                "r.18h.online",
                "-u",
                response.json()["data"]["docker_robot_account"],
                "-p",
                response.json()["data"]["docker_token"],
            ]
        )



from rockai.predictor import BasePredictor
import importlib.util
import os.path
from rockai.data_class import BaseInput
from typing import Type, Any, Union, List, Dict
import inspect
from pydantic import create_model, BaseModel, Field, ConfigDict
import enum
from typing import get_args, get_origin, Iterator
from pydantic.fields import FieldInfo
from rockai.server.types import Path as RockPath, Input, URLPath
import types
from typing_extensions import Annotated
from pathlib import Path
import re
import os

ALLOWED_INPUT_TYPES: List[Type[Any]] = [str, int, float, bool, RockPath]


def is_valid_name(string):
    # Check if the length is within the limit
    if len(string) > 140:
        return False

    # Define the regex pattern for the allowed naming convention
    pattern = r"^[a-z0-9]([-a-z0-9]*[a-z0-9])?$"

    # Use re.match to check if the string matches the pattern
    if re.match(pattern, string):
        return True
    else:
        return False


def load_class_from_file(file_path, class_name, base_class):
    """
    Dynamically load a class from a given file path, check if it is a subtype of base_class,
    and return an instance of it. Raise an error if the class is not a subtype of base_class.

    :param file_path: The path to the .py file containing the class.
    :param class_name: The name of the class to be instantiated.
    :param base_class: The class type to check against.
    :return: An instance of the class if successful.
    """
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File {file_path} does not exist.")

    module_name = os.path.splitext(os.path.basename(file_path))[0]

    spec = importlib.util.spec_from_file_location(module_name, file_path)
    if spec is None:
        raise ImportError(f"Could not load spec for {module_name} from {file_path}.")

    module = importlib.util.module_from_spec(spec)
    try:
        spec.loader.exec_module(module)
    except Exception as e:
        raise ImportError(f"Error loading module {module_name}: {e}")

    try:
        class_ = getattr(module, class_name)
    except AttributeError:
        raise AttributeError(f"Class {class_name} not found in {file_path}.")

    if not issubclass(class_, base_class):
        raise TypeError(f"{class_name} is not a subtype of {base_class.__name__}")

    instance = class_()
    return instance


def read_file_content(file_path):
    try:
        with open(file_path, "r") as file:
            content = file.read()
        return content
    except FileNotFoundError:
        return "File not found."
    except IOError:
        return "Error reading the file."


def human_readable_type_name(t: Type[Any]) -> str:
    """
    Generates a useful-for-humans label for a type. For builtin types, it's just the class name (eg "str" or "int"). For other types, it includes the module (eg "pathlib.Path" or "cog.File").

    The special case for Cog modules is because the type lives in `cog.types` internally, but just `cog` when included as a dependency.
    """
    module = t.__module__
    if module == "builtins":
        return t.__qualname__
    elif module.split(".")[0] == "cog":
        module = "cog"

    try:
        return module + "." + t.__qualname__
    except AttributeError:
        return str(t)


def readable_types_list(type_list: List[Type[Any]]) -> str:
    return ", ".join(human_readable_type_name(t) for t in type_list)


def validate_input_type(type: Type[Any], name: str) -> None:
    if type is inspect.Signature.empty:
        raise TypeError(
            f"No input type provided for parameter `{name}`. Supported input types are: {readable_types_list(ALLOWED_INPUT_TYPES)}, or a Union or List of those types."
        )
    elif type not in ALLOWED_INPUT_TYPES:
        if get_origin(type) in (Union, List, list) or (
            hasattr(types, "UnionType") and get_origin(type) is types.UnionType
        ):  # noqa: E721
            for t in get_args(type):
                validate_input_type(t, name)
        else:
            raise TypeError(
                f"Unsupported input type {human_readable_type_name(type)} for parameter `{name}`. Supported input types are: {readable_types_list(ALLOWED_INPUT_TYPES)}, or a Union or List of those types."
            )


def get_input_create_model_kwargs(signature: inspect.Signature) -> Dict[str, Any]:
    create_model_kwargs = {}

    order = 0

    for name, parameter in signature.parameters.items():
        InputType = parameter.annotation

        validate_input_type(InputType, name)

        # if no default is specified, create an empty, required input
        if parameter.default is inspect.Signature.empty:
            default = Input()
        else:
            default = parameter.default
            # If user hasn't used `Input`, then wrap it in that
            if not isinstance(default, FieldInfo):
                default = Input(default=default)

        # Fields aren't ordered, so use this pattern to ensure defined order
        # https://github.com/go-openapi/spec/pull/116

        default.json_schema_extra["x-order"] = order
        order += 1

        # Choices!
        if default.json_schema_extra.get("choices"):
            choices = default.json_schema_extra["choices"]
            # It will be passed automatically as 'enum' in the schema, so remove it as an extra field.
            del default.json_schema_extra["choices"]
            if InputType == str:

                class StringEnum(str, enum.Enum):
                    pass

                InputType = StringEnum(  # type: ignore
                    name, {value: value for value in choices}
                )
            elif InputType == int:
                InputType = enum.IntEnum(name, {str(value): value for value in choices})  # type: ignore
            else:
                raise TypeError(
                    f"The input {name} uses the option choices. Choices can only be used with str or int types."
                )

        create_model_kwargs[name] = (InputType, default)

    return create_model_kwargs


def get_input_type(predictor: BasePredictor) -> Type[BaseInput]:
    """
    Creates a Pydantic Input model from the arguments of a Predictor's predict() method.

    class Predictor(BasePredictor):
        def predict(self, text: str):
            ...

    programmatically creates a model like this:

    class Input(BaseModel):
        text: str
    """

    signature = inspect.signature(predictor.predict)

    return create_model(
        "Input",
        __config__=None,
        __base__=BaseInput,
        __module__=__name__,
        __validators__=None,
        **get_input_create_model_kwargs(signature),
    )  # type: ignore


def get_output_type(predictor: BasePredictor) -> Type[BaseModel]:
    """
    Creates a Pydantic Output model from the return type annotation of a Predictor's predict() method.
    """

    predict = predictor.predict
    signature = inspect.signature(predict)

    OutputType: Type[BaseModel]
    if signature.return_annotation is inspect.Signature.empty:
        raise TypeError(
            """You must set an output type. If your model can return multiple output types, you can explicitly set `Any` as the output type.

For example:

    from typing import Any

    def predict(
        self,
        image: Path = Input(description="Input image"),
    ) -> Any:
        ...
"""
        )
    else:
        OutputType = signature.return_annotation

    # The type that goes in the response is a list of the yielded type
    if get_origin(OutputType) is Iterator:
        # Annotated allows us to attach Field annotations to the list, which we use to mark that this is an iterator
        # https://pydantic-docs.helpmanual.io/usage/schema/#typingannotated-fields
        field = Field(**{"x-cog-array-type": "iterator"})  # type: ignore
        OutputType: Type[BaseModel] = Annotated[List[get_args(OutputType)[0]], field]  # type: ignore

    name = OutputType.__name__ if hasattr(OutputType, "__name__") else ""

    if name == "Output":
        return OutputType

    # We wrap the OutputType in an Output class to
    # ensure consistent naming of the interface in the schema.
    #
    # NOTE: If the OutputType.__name__ is "TrainingOutput" then cannot use
    # `__root__` here because this will create a reference for the Object.
    # e.g.
    #   {'title': 'Output', '$ref': '#/definitions/TrainingOutput' ... }
    #
    # And this reference may conflict with other objects at which
    # point the item will be namespaced and break our parsing. e.g.
    #   {'title': 'Output', '$ref': '#/definitions/predict_TrainingOutput' ... }
    #
    # So we work around this by inheriting from the original class rather
    # than using "__root__".
    if name == "TrainingOutput":

        class Output(OutputType):  # type: ignore
            pass

        return Output
    else:
        from pydantic import RootModel

        class Output(RootModel):
            model_config = ConfigDict(arbitrary_types_allowed=True)
            root: OutputType  # type: ignore

        return Output


def get_dependencies(obj, dependencies_type):
    """
    Checks if the given object has 'requirement_dependency' and 'system_dependency' attributes,
    ensures they are lists of strings, and returns them.

    Parameters:
        obj: The object to check.

    Returns:
        tuple: A tuple containing (requirement_dependency, system_dependency).

    Raises:
        AttributeError: If either attribute is missing.
        TypeError: If either attribute is not a list or contains non-string elements.
    """
    # Check for 'requirement_dependency' attribute
    if not hasattr(obj, dependencies_type):
        return []
    requirement_dependency = getattr(obj, f"{dependencies_type}")
    if not isinstance(requirement_dependency, list):
        raise TypeError(f"'{dependencies_type}' should be a list.")
    if not all(isinstance(item, str) for item in requirement_dependency):
        raise TypeError(f"All elements in '{dependencies_type}' should be strings.")

    return requirement_dependency


def load_predictor_class(file_path):
    file_path = Path.cwd() / file_path
    # Ensure the file exists
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"The file {file_path} does not exist.")

    # Extract module name from the file path
    module_name = os.path.splitext(os.path.basename(file_path))[0]

    # Load the module dynamically
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    # Find the class that subclasses BasePredictor
    for name, obj in inspect.getmembers(module, inspect.isclass):
        if issubclass(obj, BasePredictor) and obj is not BasePredictor:
            return obj()

    raise TypeError("No subclass of BasePredictor found in the given file.")



from rockai import (
    BasePredictor,
    Input,
    thread_limit
)
import logging

class Predictor(BasePredictor):

    def setup(self):
        #setup your model here
        self.logger = logging.getLogger()
        self.logger.setLevel("DEBUG")
        self.logger.debug("")

    #limit the number of threads runnig, the more threads the more GRAM it will be used when doing predictions
    @thread_limit(1)
    def predict(
        self, name: int = Input(description="stream total number of words")
    ) -> str:
        #start prediction
        self.logger.debug("predicting...")
        
        #return your result here
        return f"Hello, {name}"



import json
from typing import Any, Dict, List, Union
from ..server.http import create_app
import sys


def remove_title_next_to_ref(
    schema_node: Union[Dict[str, Any], List[Any]],
) -> Union[Dict[str, Any], List[Any]]:
    """
    Recursively remove 'title' from schema components that have a '$ref'.
    This function addresses a non-compliance issue in FastAPI's OpenAPI schema generation, where
    'title' fields adjacent to '$ref' fields can cause validation problems with some OpenAPI tools.
    """
    if isinstance(schema_node, dict):
        if "$ref" in schema_node and "title" in schema_node:
            del schema_node["title"]
        for _key, value in schema_node.items():
            remove_title_next_to_ref(value)
    elif isinstance(schema_node, list):  # type: ignore[reportUnnecessaryIsInstance]
        for i, item in enumerate(schema_node):
            schema_node[i] = remove_title_next_to_ref(item)
    return schema_node


def get_openapi_json(predictor_file):
    app = create_app(predictor_file, auth=None)
    return json.dumps(app.openapi(), indent=2)



import logging

# Create a custom logging handler
class PrintLoggingHandler(logging.Handler):
    def __init__(self):
        super().__init__()
        self.log_message = ""
    def emit(self, record):
        self.log_message += self.format(record)
    def get_log_msg(self):
        return self.log_message
    def clear_log_msg(self):
        self.log_message = ""

# Set up the logging configuration
def setup_logging():
    logger = logging.getLogger()  # Get the root logger
    logger.setLevel(logging.DEBUG)

    print_handler = PrintLoggingHandler()
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    print_handler.setFormatter(formatter)
    
    # Clear existing handlers to avoid duplicate logs
    # if logger.hasHandlers():
    #     logger.handlers.clear()
    
    logger.addHandler(print_handler)


import io
import mimetypes
import os
import pathlib
import shutil
import tempfile
import urllib.parse
import urllib.request
from typing import Any, Dict, Iterator, List, Optional, TypeVar, Union
import requests
from pydantic import Field, GetJsonSchemaHandler
from pydantic_core import core_schema as cs

from pydantic import GetCoreSchemaHandler
from pydantic.json_schema import JsonSchemaValue

FILENAME_ILLEGAL_CHARS = set("\u0000/")

# Linux allows files up to 255 bytes long. We enforce a slightly shorter
# filename so that there's room for prefixes added by
# tempfile.NamedTemporaryFile, etc.
FILENAME_MAX_LENGTH = 200


def Input(
    default: Any = ...,
    description: str = None,
    ge: float = None,
    le: float = None,
    min_length: int = None,
    max_length: int = None,
    regex: str = None,
    choices: List[Union[str, int]] = None,
) -> Any:
    """Input is similar to pydantic.Field, but doesn't require a default value to be the first argument."""
    return Field(
        default,
        description=description,
        ge=ge,
        le=le,
        min_length=min_length,
        max_length=max_length,
        regex=regex,
        choices=choices,
    )


class File(io.IOBase):
    validate_always = True

    @classmethod
    def __get_validators__(cls) -> Iterator[Any]:
        yield cls.validate

    @classmethod
    def validate(cls, value: Any) -> io.IOBase:
        if isinstance(value, io.IOBase):
            return value

        parsed_url = urllib.parse.urlparse(value)
        if parsed_url.scheme == "data":
            res = urllib.request.urlopen(value)  # noqa: S310
            return io.BytesIO(res.read())
        elif parsed_url.scheme == "http" or parsed_url.scheme == "https":
            return URLFile(value)
        else:
            raise ValueError(
                f"'{parsed_url.scheme}' is not a valid URL scheme. 'data', 'http', or 'https' is supported."
            )

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> cs.CoreSchema:
        return cs.str_schema()

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: cs.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        json_schema = handler(core_schema)
        json_schema = handler.resolve_ref_schema(json_schema)
        json_schema["format"] = "uri"
        json_schema["type"] = "string"
        return json_schema


class Path(pathlib.PosixPath):
    validate_always = True

    @classmethod
    def __get_validators__(cls) -> Iterator[Any]:
        yield cls.validate

    @classmethod
    def validate(cls, value: Any) -> pathlib.Path:
        if isinstance(value, pathlib.Path):
            return value

        return URLPath(
            source=value,
            filename=get_filename(value),
            fileobj=File.validate(value),
        )

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> cs.CoreSchema:
        return cs.str_schema()

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: cs.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        json_schema = handler(core_schema)
        json_schema = handler.resolve_ref_schema(json_schema)
        json_schema["format"] = "uri"
        json_schema["type"] = "string"
        return json_schema


class URLPath(pathlib.PosixPath):
    """
    URLPath is a nasty hack to ensure that we can defer the downloading of a
    URL passed as a path until later in prediction dispatch.

    It subclasses pathlib.PosixPath only so that it can pass isinstance(_,
    pathlib.Path) checks.
    """

    _path: Optional[Path]

    def __init__(self, *, source: str, filename: str, fileobj: io.IOBase) -> None:
        self.source = source
        self.filename = filename
        self.fileobj = fileobj

        self._path = None

    def convert(self) -> Path:
        if self._path is None:
            dest = tempfile.NamedTemporaryFile(suffix=self.filename, delete=False)
            shutil.copyfileobj(self.fileobj, dest)
            self._path = Path(dest.name)
        return self._path

    def unlink(self, missing_ok: bool = False) -> None:
        if self._path:
            # TODO: use unlink(missing_ok=...) when we drop Python 3.7 support.
            try:
                self._path.unlink()
            except FileNotFoundError:
                if not missing_ok:
                    raise

    def __str__(self) -> str:
        # FastAPI's jsonable_encoder will encode subclasses of pathlib.Path by
        # calling str() on them
        return self.source


class URLFile(io.IOBase):
    """
    URLFile is a proxy object for a :class:`urllib3.response.HTTPResponse`
    object that is created lazily. It's a file-like object constructed from a
    URL that can survive pickling/unpickling.
    """

    __slots__ = ("__target__", "__url__")

    def __init__(self, url: str) -> None:
        object.__setattr__(self, "__url__", url)

    # We provide __getstate__ and __setstate__ explicitly to ensure that the
    # object is always picklable.
    def __getstate__(self) -> Dict[str, Any]:
        return {"url": object.__getattribute__(self, "__url__")}

    def __setstate__(self, state: Dict[str, Any]) -> None:
        object.__setattr__(self, "__url__", state["url"])

    # Proxy getattr/setattr/delattr through to the response object.
    def __setattr__(self, name: str, value: Any) -> None:
        if hasattr(type(self), name):
            object.__setattr__(self, name, value)
        else:
            setattr(self.__wrapped__, name, value)

    def __getattr__(self, name: str) -> Any:
        if name in ("__target__", "__wrapped__", "__url__"):
            raise AttributeError(name)
        else:
            return getattr(self.__wrapped__, name)

    def __delattr__(self, name: str) -> None:
        if hasattr(type(self), name):
            object.__delattr__(self, name)
        else:
            delattr(self.__wrapped__, name)

    # Luckily the only dunder method on HTTPResponse is __iter__
    def __iter__(self) -> Iterator[bytes]:
        return iter(self.__wrapped__)

    @property
    def __wrapped__(self) -> Any:
        try:
            return object.__getattribute__(self, "__target__")
        except AttributeError:
            url = object.__getattribute__(self, "__url__")
            resp = requests.get(url, stream=True)
            resp.raise_for_status()
            resp.raw.decode_content = True
            object.__setattr__(self, "__target__", resp.raw)
            return resp.raw

    def __repr__(self) -> str:
        try:
            target = object.__getattribute__(self, "__target__")
        except AttributeError:
            return "<{} at 0x{:x} for {!r}>".format(
                type(self).__name__, id(self), object.__getattribute__(self, "__url__")
            )
        else:
            return f"<{type(self).__name__} at 0x{id(self):x} wrapping {target!r}>"


def get_filename(url: str) -> str:
    parsed_url = urllib.parse.urlparse(url)

    if parsed_url.scheme == "data":
        resp = urllib.request.urlopen(url)  # noqa: S310
        mime_type = resp.headers.get_content_type()
        extension = mimetypes.guess_extension(mime_type)
        if extension is None:
            return "file"
        return "file" + extension

    basename = os.path.basename(parsed_url.path)
    basename = urllib.parse.unquote_plus(basename)

    # If the filename is too long, we truncate it (appending '~' to denote the
    # truncation) while preserving the file extension.
    # - truncate it
    # - append a tilde
    # - preserve the file extension
    if _len_bytes(basename) > FILENAME_MAX_LENGTH:
        basename = _truncate_filename_bytes(basename, length=FILENAME_MAX_LENGTH)

    for c in FILENAME_ILLEGAL_CHARS:
        basename = basename.replace(c, "_")

    return basename


Item = TypeVar("Item")


class ConcatenateIterator(Iterator[Item]):
    @classmethod
    def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
        """Defines what this type should be in openapi.json"""
        field_schema.pop("allOf", None)

        field_schema.update(
            {
                "type": "array",
                "items": {"type": "string"},
                "x-cog-array-type": "iterator",
                "x-cog-array-display": "concatenate",
            }
        )

    @classmethod
    def __get_validators__(cls) -> Iterator[Any]:
        yield cls.validate

    @classmethod
    def validate(cls, value: Iterator[Any]) -> Iterator[Any]:
        return value

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> cs.CoreSchema:
        return cs.list_schema()

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: cs.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        json_schema = handler(core_schema)
        json_schema = handler.resolve_ref_schema(json_schema)
        json_schema.pop("allOf", None)
        json_schema["type"] = "array"
        json_schema["items"] = {"type": "string"}
        json_schema["x-cog-array-type"] = "iterator"
        json_schema["x-cog-array-display"] = "concatenate"
        return json_schema


def _len_bytes(s: str, encoding: str = "utf-8") -> int:
    return len(s.encode(encoding))


def _truncate_filename_bytes(s: str, length: int, encoding: str = "utf-8") -> str:
    """
    Truncate a filename to at most `length` bytes, preserving file extension
    and avoiding text encoding corruption from truncation.
    """
    root, ext = os.path.splitext(s.encode(encoding))
    root = root[: length - len(ext) - 1]
    return root.decode(encoding, "ignore") + "~" + ext.decode(encoding)



torch_version_list = [
    {
        "version": "2.4.0",
        "install": "pip install torch==2.4.0 torchvision==0.19.0 torchaudio==2.4.0 --index-url https://download.pytorch.org/whl/cu118 --extra-index-url https://pypi.tuna.tsinghua.edu.cn/simple",
        "image": "nvidia/cuda:11.8.0-cudnn8-devel-ubuntu22.04",
    },
    {
        "version": "2.3.1",
        "install": "pip install torch==2.3.1 torchvision==0.18.1 torchaudio==2.3.1 --index-url https://download.pytorch.org/whl/cu118 --extra-index-url https://pypi.tuna.tsinghua.edu.cn/simple",
        "image": "nvidia/cuda:11.8.0-cudnn8-devel-ubuntu22.04",
    },
    {
        "version": "2.3.0",
        "install": "pip install torch==2.3.0 torchvision==0.18.0 torchaudio==2.3.0 --index-url https://download.pytorch.org/whl/cu118 --extra-index-url https://pypi.tuna.tsinghua.edu.cn/simple",
        "image": "nvidia/cuda:11.8.0-cudnn8-devel-ubuntu22.04",
    },
    {
        "version": "2.2.2",
        "install": "pip install torch==2.2.2 torchvision==0.17.2 torchaudio==2.2.2 --index-url https://download.pytorch.org/whl/cu118 --extra-index-url https://pypi.tuna.tsinghua.edu.cn/simple",
        "image": "nvidia/cuda:11.8.0-cudnn8-devel-ubuntu22.04",
    },
    {
        "version": "2.2.1",
        "install": "pip install torch==2.2.1 torchvision==0.17.1 torchaudio==2.2.1 --index-url https://download.pytorch.org/whl/cu118 --extra-index-url https://pypi.tuna.tsinghua.edu.cn/simple",
        "image": "nvidia/cuda:11.8.0-cudnn8-devel-ubuntu22.04",
    },
    {
        "version": "2.2.0",
        "install": "pip install torch==2.2.0 torchvision==0.17.0 torchaudio==2.2.0 --index-url https://download.pytorch.org/whl/cu118 --extra-index-url https://pypi.tuna.tsinghua.edu.cn/simple",
        "image": "nvidia/cuda:11.8.0-cudnn8-devel-ubuntu22.04",
    },
    {
        "version": "2.1.2",
        "install": "pip install torch==2.1.2 torchvision==0.16.2 torchaudio==2.1.2 --index-url https://download.pytorch.org/whl/cu118 --extra-index-url https://pypi.tuna.tsinghua.edu.cn/simple",
        "image": "nvidia/cuda:11.8.0-cudnn8-devel-ubuntu22.04",
    },
    {
        "version": "2.1.1",
        "install": "pip install torch==2.1.1 torchvision==0.16.1 torchaudio==2.1.1 --index-url https://download.pytorch.org/whl/cu118 --extra-index-url https://pypi.tuna.tsinghua.edu.cn/simple",
        "image": "nvidia/cuda:11.8.0-cudnn8-devel-ubuntu22.04",
    },
    {
        "version": "2.1.0",
        "install": "pip install torch==2.1.0 torchvision==0.16.0 torchaudio==2.1.0 --index-url https://download.pytorch.org/whl/cu118 --extra-index-url https://pypi.tuna.tsinghua.edu.cn/simple",
        "image": "nvidia/cuda:11.8.0-cudnn8-devel-ubuntu22.04",
    },
    {
        "version": "2.0.1",
        "install": "pip install torch==2.0.1 torchvision==0.15.2 torchaudio==2.0.2 --index-url https://download.pytorch.org/whl/cu118 --extra-index-url https://pypi.tuna.tsinghua.edu.cn/simple",
        "image": "nvidia/cuda:11.8.0-cudnn8-devel-ubuntu22.04",
    },
    {
        "version": "2.0.0",
        "install": "pip install torch==2.0.0 torchvision==0.15.1 torchaudio==2.0.1 --index-url https://download.pytorch.org/whl/cu118 --extra-index-url https://pypi.tuna.tsinghua.edu.cn/simple",
        "image": "nvidia/cuda:11.8.0-cudnn8-devel-ubuntu22.04",
    },
    {
        "version": "1.13.1",
        "install": "pip install torch==1.13.1+cu117 torchvision==0.14.1+cu117 torchaudio==0.13.1 --extra-index-url https://download.pytorch.org/whl/cu117 --index-url https://pypi.tuna.tsinghua.edu.cn/simple",
        "image": "nvidia/cuda:11.7.1-cudnn8-devel-ubuntu22.04",
    },
    {
        "version": "1.13.0",
        "install": "pip install torch==1.13.0+cu117 torchvision==0.14.0+cu117 torchaudio==0.13.0 --extra-index-url https://download.pytorch.org/whl/cu117 --index-url https://pypi.tuna.tsinghua.edu.cn/simple",
        "image": "nvidia/cuda:11.7.1-cudnn8-devel-ubuntu22.04",
    },
    {
        "version": "1.12.1",
        "install": "pip install torch==1.12.1+cu116 torchvision==0.13.1+cu116 torchaudio==0.12.1 --extra-index-url https://download.pytorch.org/whl/cu116 --index-url https://pypi.tuna.tsinghua.edu.cn/simple",
        "image": "nvidia/cuda:11.6.1-cudnn8-devel-ubuntu20.04",
    },
    {
        "version": "1.12.0",
        "install": "pip install torch==1.12.0+cu116 torchvision==0.13.0+cu116 torchaudio==0.12.0 --extra-index-url https://download.pytorch.org/whl/cu116 --index-url https://pypi.tuna.tsinghua.edu.cn/simple",
        "image": "nvidia/cuda:11.6.1-cudnn8-devel-ubuntu20.04",
    },
]



import os
from typing import Iterator, Dict, Optional, Any, List
import requests
import time
import httpx
from httpx_sse import connect_sse
import logging
import asyncio
from aiohttp_sse_client import client as sse_client
import aiohttp
from rockai.server.types import Path, Input, ConcatenateIterator
from pydantic import BaseModel
from rockai.predictor import BasePredictor
import threading
from functools import wraps

logging.basicConfig(level=logging.DEBUG)

__all__ = [
    "BaseModel",
    "BasePredictor",
    "ConcatenateIterator",
    "Input",
    "Path",
    "thread_limit",
]


def thread_limit(limit=1):
    semaphore = threading.BoundedSemaphore(limit)

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            with semaphore:
                return func(*args, **kwargs)

        return wrapper

    return decorator


# client for rock AI
class Client:

    def __init__(self, api_token: str = None):
        self.api_token = (
            api_token if api_token is not None else os.environ.get("ROCK_API_TOKEN")
        )

        # Local testing
        # self.predict_url = "http://localhost:8000/v1/predictions"
        # self.get_url = "http://localhost:8000/v1/predictions/{}"
        # self.cancel_url = "http://localhost:8000/v1/predictions/{}/cancel"

        # Production
        self.predict_url = "https://api.rockai.online/v1/predictions"
        self.get_url = "https://api.rockai.online/v1/predictions/{}"
        self.cancel_url = "https://api.rockai.online/v1/predictions/{}/cancel"

        # Dev
        # self.predict_url = "https://api-dev.rockai.online/v1/predictions"
        # self.get_url = "https://api-dev.rockai.online/v1/predictions/{}"
        # self.cancel_url = "https://api-dev.rockai.online/v1/predictions/{}/cancel"

    # Create a new prediction
    def create(
        self,
        model: str,
        version: Optional[str] = None,
        input: Optional[Dict[str, Any]] = None,
        webhook: Optional[str] = None,
        webhook_events_filter: Optional[List[str]] = None,
    ):
        url = self.predict_url

        payload = {"input": input, "model": model}

        if webhook is not None:
            payload["webhook"] = webhook
        if webhook_events_filter is not None:
            payload["webhook_events_filter"] = webhook_events_filter
        if version is not None:
            payload["version"] = version
        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer {}".format(self.api_token),
        }

        response = requests.post(url=url, headers=headers, json=payload)
        response.raise_for_status()
        return response.json()

    async def create_async(
        self,
        model: str,
        version: Optional[str],
        input: Optional[Dict[str, Any]] = None,
        webhook: Optional[str] = None,
        webhook_events_filter: Optional[List[str]] = None,
    ):
        url = self.predict_url

        payload = {"input": input, "model": model}
        if webhook is not None:
            payload["webhook"] = webhook
        if webhook_events_filter is not None:
            payload["webhook_events_filter"] = webhook_events_filter
        if "version" != None:
            payload["version"] = version
        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer {}".format(self.api_token),
        }

        async with httpx.AsyncClient() as client:
            response = await client.post(url=url, headers=headers, json=payload)
            response.raise_for_status()
            return response.json()

    # Get a prediction result by ID
    def get(self, id: str):
        headers = {"Authorization": f"Bearer {self.api_token}"}
        response = requests.get(self.get_url.format(id), headers=headers)
        response.raise_for_status()  # Raise an exception for HTTP errors
        result = response.json()
        logging.info(result)
        return result

    # Get a prediction result by ID (async)
    async def get_async(self, id: str):
        headers = {"Authorization": f"Bearer {self.api_token}"}

        async with aiohttp.ClientSession() as session:
            async with session.get(
                self.get_url.format(id), headers=headers
            ) as response:
                response.raise_for_status()  # Raise an exception for HTTP errors
                result = await response.json()
                logging.info(result)
                return result

    # Stream output from a model, if model supports streaming like LLama3
    def stream(
        self,
        model: str,
        version: Optional[str] = None,
        input: Optional[Dict[str, Any]] = None,
    ) -> Iterator:
        url = self.predict_url

        payload = {"input": input, "stream": True, "model": model}
        if version is not None:
            payload["version"] = version
        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer {}".format(self.api_token),
        }

        response = requests.post(url=url, headers=headers, json=payload)
        response.raise_for_status()
        create_result = response.json()
        if (
            response.status_code == 200
            or response.status_code == 201
            and "stream" in create_result["data"]["urls"]
        ):
            headers["Accept"] = "text/event-stream"
            headers["Cache-Control"] = "no-store"

            with httpx.Client() as client:
                with connect_sse(
                    client, create_result["data"]["urls"]["stream"]
                ) as event_source:
                    try:
                        for sse in event_source.iter_sse():
                            yield sse.data
                    except Exception as e:
                        logging.error(str(e))

    # Run a model and return the result from the model
    def run(
        self,
        model: str,
        version: Optional[str] = None,
        input: Optional[Dict[str, Any]] = None,
        webhook: Optional[str] = None,
        webhook_events_filter: Optional[List[str]] = None,
    ) -> Any:
        url = self.predict_url

        payload = {"input": input, "model": model}
        if version is not None:
            payload["version"] = version
        if webhook is not None:
            payload["webhook"] = webhook
        if webhook_events_filter is not None:
            payload["webhook_events_filter"] = webhook_events_filter

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer {}".format(self.api_token),
        }

        response = requests.post(url=url, headers=headers, json=payload)
        response.raise_for_status()
        create_result = response.json()
        logging.info(create_result)
        if response.status_code == 200 or response.status_code == 201:
            while True:
                get_resp = requests.get(
                    url=self.get_url.format(create_result["id"]),
                    headers=headers,
                )
                get_resp.raise_for_status()
                get_result = get_resp.json()
                if (
                    get_result["status"] == "processing"
                    or get_result["status"] == "starting"
                ):
                    time.sleep(1)
                    continue
                else:
                    return get_result

    # Stream output from a model, if model supports streaming like LLama3 (async version)
    async def stream_async(
        self,
        model: str,
        version: Optional[str] = None,
        input: Optional[Dict[str, Any]] = None,
    ):
        url = self.predict_url

        payload = {"input": input, "stream": True, "model": model}
        if version is not None:
            payload["version"] = version
        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer {}".format(self.api_token),
        }
        async with httpx.AsyncClient() as client:
            response = await client.post(url=url, headers=headers, json=payload)
            response.raise_for_status()
            create_result = response.json()
            if "stream" in create_result["urls"]:

                logging.info(create_result["urls"])
                async with sse_client.EventSource(
                    create_result["urls"]["stream"]
                ) as event_source:

                    async for event in event_source:
                        yield event.data
                        if event.type == "done" and event.message == "done":
                            return

    # Run a model and return the result from the model (Async version)
    async def run_async(
        self,
        model: str,
        version: Optional[str] = None,
        input: Optional[Dict[str, Any]] = None,
        webhook: Optional[str] = None,
        webhook_events_filter: Optional[List[str]] = None,
    ):
        url = self.predict_url

        payload = {"input": input, "model": model}
        if webhook is not None:
            payload["webhook"] = webhook
        if webhook_events_filter is not None:
            payload["webhook_events_filter"] = webhook_events_filter
        if version is not None:
            payload["version"] = version
        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer {}".format(self.api_token),
        }

        async with httpx.AsyncClient() as client:
            response = await client.post(url=url, headers=headers, json=payload)
            response.raise_for_status()
            create_result = response.json()
            while True:
                get_resp = await client.get(
                    url=self.get_url.format(create_result["id"]),
                    headers=headers,
                )
                get_result = get_resp.json()
                if (
                    get_result["status"] == "processing"
                    or get_result["status"] == "starting"
                ):
                    await asyncio.sleep(1)
                    continue
                else:
                    return get_result



from fastapi import HTTPException, Security
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
import logging

logging.basicConfig(level=logging.DEBUG)


class AuthHandler:
    def __init__(self, auth_token):
        self.auth_token = auth_token

    is_using_auth = False
    security = HTTPBearer()
    optional_security = HTTPBearer(auto_error=False)

    def __check_access_token__(self, token) -> str:
        if token == self.auth_token:
            return "success"
        else:
            raise HTTPException(status_code=401, detail="invalid token")

    def auth_wrapper(
        self,
        auth: HTTPAuthorizationCredentials = Security(
            security if is_using_auth else optional_security
        ),
    ):
        if self.is_using_auth:
            if auth is None:
                raise HTTPException(status_code=401, detail="not authenticated")
            return self.__check_access_token__(auth.credentials)
        else:
            return "success"
